"""
Absfuyu: Data Analysis
----------------------
Data Analyst DataFrame - Base/Core

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["DataAnalystDataFrameBase", "SplittedDF", "CityData"]


# Library
# ---------------------------------------------------------------------------
import random
from collections import deque
from typing import ClassVar, Literal, NamedTuple

import pandas as pd


# Class
# ---------------------------------------------------------------------------
class DataAnalystDataFrameBase(pd.DataFrame):
    """
    Data Analyst ``pd.DataFrame`` - Base

    Set class variable ``_DADF_INCLUDE`` to ``False`` to exclude from ``DADF_METHODS``
    """

    # Custom attribute
    _DADF_INCLUDE: ClassVar[bool] = True  # Include in DADF_METHODS
    DADF_METHODS: ClassVar[dict[str, list[str]]] = {}

    def __init_subclass__(cls, *args, **kwargs) -> None:
        """
        This create a dictionary with:
        - key   (str)      : Subclass
        - value (list[str]): List of available methods
        """
        super().__init_subclass__(*args, **kwargs)

        if cls._DADF_INCLUDE and not any(
            [x.endswith(cls.__name__) for x in cls.DADF_METHODS.keys()]
        ):
            # if not any([x.endswith(cls.__name__) for x in cls.DADF_METHODS.keys()]):
            methods_list: list[str] = [
                k for k, v in cls.__dict__.items() if callable(v)
            ]
            if len(methods_list) > 0:
                name = f"{cls.__module__}.{cls.__name__}"
                cls.DADF_METHODS.update({name: sorted(methods_list)})


class SplittedDF(NamedTuple):
    """
    DataFrame splitted into contains
    missing values only and vice versa

    Parameters
    ----------
    df : DataFrame
        DataFrame without missing values

    df_na : DataFrame
        DataFrame with missing values only
    """

    df: pd.DataFrame
    df_na: pd.DataFrame

    @staticmethod
    def concat_df(
        df_list: list[pd.DataFrame], join: Literal["inner", "outer"] = "inner"
    ) -> pd.DataFrame:
        """
        Concat the list of DataFrame (static method)

        Parameters
        ----------
        df_list : list[DataFrame]
            A sequence of DataFrame

        join : str
            Join type
            (Default: ``"inner"``)

        Returns
        -------
        DataFrame
            Joined DataFrame
        """
        df: pd.DataFrame = pd.concat(df_list, axis=0, join=join).reset_index()
        df.drop(columns=["index"], inplace=True)
        return df

    def concat(self, join: Literal["inner", "outer"] = "inner") -> pd.DataFrame:
        """
        Concat the splitted DataFrame

        Parameters
        ----------
        join : str
            Join type
            (Default: ``"inner"``)

        Returns
        -------
        DataFrame
            Joined DataFrame
        """
        return self.concat_df(self, join=join)  # type: ignore

    @staticmethod
    def divide_dataframe(df: pd.DataFrame, by_column: str) -> list[pd.DataFrame]:
        """
        Divide DataFrame into a list of DataFrame

        Parameters
        ----------
        df : DataFrame
            DataFrame

        by_column : str
            By which column

        Returns
        -------
        list[DataFrame]
            Splitted DataFrame
        """
        divided = [x for _, x in df.groupby(by_column)]
        return divided


class CityData(NamedTuple):
    """
    Parameters
    ----------
    city : str
        City name

    region : str
        Region of the city

    area : str
        Area of the region
    """

    city: str
    region: str
    area: str

    @staticmethod
    def _sample_city_data(size: int = 100):
        """
        Generate sample city data (testing purpose)
        """
        sample_range = 10 ** len(str(size))

        # Serial list
        serials: list[str] = []
        while len(serials) != size:  # Unique serial
            serial = random.randint(0, sample_range - 1)
            serial = str(serial).rjust(len(str(size)), "0")  # type: ignore
            if serial not in serials:  # type: ignore
                serials.append(serial)  # type: ignore

        ss2 = deque(serials[: int(len(serials) / 2)])  # Cut half for region
        ss2.rotate(random.randrange(1, 5))
        [ss2.extend(ss2) for _ in range(2)]  # type: ignore # Extend back

        ss3 = deque(serials[: int(len(serials) / 4)])  # Cut forth for area
        ss3.rotate(random.randrange(1, 5))
        [ss3.extend(ss3) for _ in range(4)]  # type: ignore # Extend back

        serials = ["city_" + x for x in serials]
        ss2 = ["region_" + x for x in ss2]  # type: ignore
        ss3 = ["area_" + x for x in ss3]  # type: ignore

        ss = list(zip(serials, ss2, ss3))  # Zip back
        out = list(map(CityData._make, ss))

        return out



"""
Test: Beautiful

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

import pytest

try:  # [beautiful] feature
    import rich  # type: ignore
    from rich.console import Group
except ImportError:
    rich = pytest.importorskip("rich")

from absfuyu.extra.beautiful import BeautifulOutput


# Sample function to decorate
@BeautifulOutput(layout=1)
def sample_function(x: int, y: int) -> int:
    return x + y


class TestBeautifulOutput:
    def test_basic_functionality(self) -> None:
        """Test basic functionality of the BeautifulOutput decorator."""
        result = sample_function(2, 3)
        assert isinstance(result, Group)

    def test_invalid_layout(self) -> None:
        """Test that an invalid layout defaults to layout 1."""
        decorator = BeautifulOutput(layout=99)  # Invalid layout
        decorated_function = decorator(sample_function)

        output = decorated_function(2, 3)

        # Check if it defaults to layout 1
        assert isinstance(output, Group)

    def test_memory_usage_measurement(self) -> None:
        """Test that memory usage is measured correctly."""

        @BeautifulOutput()
        def memory_intensive_function(n: int) -> list:
            return [i for i in range(n)]

        output = memory_intensive_function(10000)

        # Ensure that the function runs without errors and returns a list
        assert isinstance(output, Group)

    def test_error_handling(self) -> None:
        """Test error handling when an exception occurs in the wrapped function."""

        @BeautifulOutput()
        def error_prone_function():
            raise ValueError("An error occurred!")

        with pytest.raises(ValueError) as excinfo:
            error_prone_function()

        assert "An error occurred!" in str(excinfo.value)



"""
Absfuyu: XML
------------
XML Tool [W.I.P]

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["XML2Dict"]


# Library
# ---------------------------------------------------------------------------
from pathlib import Path
from typing import Any, Self

from absfuyu.core.baseclass import GetClassMembersMixin

XML_MODE = False

try:
    import xmltodict
except ImportError:
    from subprocess import run

    from absfuyu.config import ABSFUYU_CONFIG

    if ABSFUYU_CONFIG._get_setting("auto-install-extra").value:  # type: ignore
        cmd = "python -m pip install -U absfuyu[full]".split()
        run(cmd)
    else:
        raise SystemExit("This feature is in absfuyu[full] package")  # noqa: B904
else:
    XML_MODE = True


# Class
# ---------------------------------------------------------------------------
class XML2Dict(GetClassMembersMixin):
    """
    Automatically convert .xml file to dict


    Usage
    -----
    >>> XML2Dict(<path>)
    >>> XML2Dict.parsed_data
    """

    def __init__(self, text: str) -> None:
        """
        Automatically convert .xml file to dict

        Parameters
        ----------
        text : str
            ``.xml`` format text
        """
        self._text = text

        self.parsed_data: dict[str, Any] = self._parse_xml()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def _parse_xml(self) -> dict[str, Any]:
        """Convert xml to dict"""
        return xmltodict.parse(self._text, encoding="utf-8")  # type: ignore

    @classmethod
    def from_path(cls, xml_path: str | Path) -> Self:
        """
        Convert from .xml file path

        Parameters
        ----------
        xml_path : str | Path
            Path to .xml file

        Returns
        -------
        Self
            xml to dict
        """
        with Path(xml_path).open("r", encoding="utf-8") as f:
            dat = "\n".join(f.readlines())
        return cls(dat)



"""
Absfuyu: Fun
------------
Some fun or weird stuff

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["zodiac_sign", "happy_new_year", "human_year_to_dog_year"]


# Library
# ---------------------------------------------------------------------------
from datetime import date

from absfuyu.core import deprecated, versionadded, versionchanged
from absfuyu.logger import logger
from absfuyu.tools.shutdownizer import ShutDownizer
from absfuyu.util.lunar import LunarCalendar


# Function
# ---------------------------------------------------------------------------
@versionchanged("5.0.0", reason="Updated functionality")
def zodiac_sign(day: int, month: int, zodiac13: bool = False) -> str:
    """
    Calculate zodiac sign

    Parameters
    ----------
    day : int
        Day (in range [1, 31])

    month : int
        Month (in range [1, 12])

    zodiac13 : bool
        13 zodiacs mode (Default: ``False``)

    Returns
    -------
    str
        Zodiac sign
    """

    # Condition check
    if not 1 <= month <= 12 or not 1 <= day <= 31:
        raise ValueError("Value out of range")

    if zodiac13:
        zodiac_dates = {
            "Capricorn": ((1, 21), (2, 16)),
            "Aquarius": ((2, 17), (3, 11)),
            "Pisces": ((3, 12), (4, 18)),
            "Aries": ((4, 19), (5, 13)),
            "Taurus": ((5, 14), (6, 21)),
            "Gemini": ((6, 22), (7, 20)),
            "Cancer": ((7, 21), (8, 10)),
            "Leo": ((8, 11), (9, 16)),
            "Virgo": ((9, 17), (10, 30)),
            "Libra": ((10, 31), (11, 23)),
            "Scorpio": ((11, 24), (11, 29)),
            "Ophiuchus": ((11, 30), (12, 17)),
            "Sagittarius": ((12, 18), (1, 20)),
        }
    else:
        zodiac_dates = {
            "Capricorn (E)": ((12, 22), (1, 19)),
            "Aquarius (A)": ((1, 20), (2, 18)),
            "Pisces (W)": ((2, 19), (3, 20)),
            "Aries (F)": ((3, 21), (4, 19)),
            "Taurus (E)": ((4, 20), (5, 20)),
            "Gemini (A)": ((5, 21), (6, 20)),
            "Cancer (W)": ((6, 21), (7, 22)),
            "Leo (F)": ((7, 23), (8, 22)),
            "Virgo (E)": ((8, 23), (9, 22)),
            "Libra (A)": ((9, 23), (10, 22)),
            "Scorpio (W)": ((10, 23), (11, 21)),
            "Sagittarius (F)": ((11, 22), (12, 21)),
        }

    result = ""
    for sign, (start_date, end_date) in zodiac_dates.items():
        start_month, start_day = start_date
        end_month, end_day = end_date

        if (month == start_month and day >= start_day) or (
            month == end_month and day <= end_day
        ):
            result = sign
            break
    return result


@deprecated("5.0.0", reason="API shutdown, will be removed later")
def im_bored() -> str:
    """
    Get random activity from ``boredapi`` website

    Returns
    -------
    str
        Random activity
    """
    raise SystemExit("API shuted down")
    # try:
    #     api = APIRequest("https://www.boredapi.com/api/activity")
    #     return api.fetch_data_only().json()["activity"]  # type: ignore
    # except Exception:
    #     return "FAILED"


# For new year only
def happy_new_year(forced: bool = False, include_lunar: bool = False) -> None:
    """
    Only occurs on 01/01 every year

    Parameters
    ----------
    forced : bool
        Shutdown ASAP (Default: ``False``)

    include_lunar : bool
        Include Lunar New Year (Default: ``False``)
    """

    if forced:
        return ShutDownizer().shutdown()

    today = date.today()
    m = today.month
    d = today.day
    solar_new_year = m == 1 and d == 1
    logger.debug(f"Solar: {today}")

    if include_lunar:
        lunar = LunarCalendar.now().date
        lunar_new_year = lunar.month == 1 and lunar.day == 1
        logger.debug(f"Lunar: {lunar}")
    else:
        lunar_new_year = False

    if solar_new_year or lunar_new_year:
        print("Happy New Year! You should take rest now.")
        return ShutDownizer().shutdown()
    else:
        raise SystemExit("The time has not come yet")


@versionadded("5.0.0")
def human_year_to_dog_year(
    human_year: int | float, is_small: bool = True
) -> int | float:
    """
    Convert human's year to dog's year

    General Guidelines:
    1. First Year: The first year of a dog's life
    is roughly equal to 15 human years.
    2. Second Year: The second year adds about 9 human years,
    making a 2-year-old dog equivalent to about 24 human years.
    3. Subsequent Years: After the second year,
    each additional yeartypically equals about 4-5 human years,
    depending on the dog's size and breed. (Small-4, Large-5)

    Parameters
    ----------
    human_year : int | float
        Dog's age in human year

    is_small : bool, optional
        Is the dog small or not, by default ``True``

    Returns
    -------
    int | float
        Dog's year/age

    Raises
    ------
    ValueError
        When ``human_year`` < 0
    """

    if human_year < 0:
        raise ValueError("Value must be positive")
    if human_year <= 1:
        return human_year * 15
    elif human_year <= 2:
        return 15 + (human_year - 1) * 9
    else:
        if is_small:
            return 24 + (human_year - 2) * 4
        else:
            return 24 + (human_year - 2) * 5



"""
Absufyu: API
------------
Fetch data stuff

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = [
    "APIRequest",
    "ping_windows",
]


# Library
# ---------------------------------------------------------------------------
import json
import re
import subprocess
from pathlib import Path
from typing import NamedTuple

import requests

from absfuyu.core.baseclass import BaseClass
from absfuyu.core.docstring import versionadded, versionchanged
from absfuyu.logger import logger


# Function
# ---------------------------------------------------------------------------
class PingResult(NamedTuple):
    """
    Ping result

    Parameters
    ----------
    host : str
        Host name/IP

    result : str
        Ping result in ms
    """

    host: str
    result: str


@versionchanged("3.4.0", reason="Updated functionality")
@versionadded("2.5.0")
def ping_windows(host: list[str], ping_count: int = 3) -> list[PingResult]:
    """
    Ping web

    Parameters
    ----------
    host : list[str]
        List of host to ping

    ping_count : int
        Number of time to ping to take average
        (Default: ``3``)

    Returns
    -------
    list
        List of host with pinged value


    Example:
    --------
    >>> ping_windows(["1.1.1.1", "google.com"])
    ['1.1.1.1 : xxms', 'google.com : xxms']
    """
    out: list[PingResult] = []

    for ip in host:
        output = subprocess.run(
            f"ping {ip.strip()} -n {ping_count}",
            encoding="utf-8",
            capture_output=True,
            text=True,
        )
        logger.debug(output)

        data: str = "".join(output.stdout)
        res = re.findall(r"Average = (.*)", data)
        if res:
            out.append(PingResult(ip, res[0]))
        else:
            out.append(PingResult(ip, "FAILED"))

    return out


# Class
# ---------------------------------------------------------------------------
class APIRequest(BaseClass):
    """
    API data with cache feature

    Parameters
    ----------
    api_url : str
        API link

    encoding : str | None, optional
        Data encoding, by default ``"utf-8"``
    """

    def __init__(
        self,
        api_url: str,
        *,
        encoding: str | None = "utf-8",
    ) -> None:
        """
        Create APIRequest instance

        Parameters
        ----------
        api_url : str
            API link

        encoding : str | None, optional
            Data encoding, by default ``"utf-8"``
        """

        self.url = api_url
        self.encoding = encoding

    def fetch_data(self, *, update: bool = False, json_cache: str | Path):
        """
        Fetch data from an API then cache it for later use

        Parameters
        ----------
        update :
            Refresh the cache when ``True``
            (Default: ``False``)

        json_cache : Path | str
            Name of the cache

        Returns
        -------
        Any
           Data

        None
            No data fetched/unable to fetch data
        """
        if update:
            json_data = None
        else:
            try:
                with open(json_cache, "r", encoding=self.encoding) as file:
                    json_data = json.load(file)
                    logger.debug("Fetched data from local cache!")
            except (FileNotFoundError, json.JSONDecodeError) as e:
                logger.debug(f"No local cache found... ({e})")
                json_data = None

        if json_data is None:
            logger.debug("Fetching new json data... (Creating local cache)")
            try:
                json_data = requests.get(self.url).json()
                with open(json_cache, "w", encoding=self.encoding) as file:
                    json.dump(json_data, file, indent=2)
            except FileNotFoundError as e:
                logger.error(f"Can't create cache due to Path error - {e}")

        return json_data

    def fetch_data_only(self) -> requests.Response:
        """
        Fetch data without cache

        Returns
        -------
        Response
            ``requests.Response``
        """
        return requests.get(self.url)



"""
Absfuyu: Core
-------------
Sphinx docstring decorator

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module Package
# ---------------------------------------------------------------------------
__all__ = [
    "SphinxDocstring",
    "SphinxDocstringMode",
    "versionadded",
    "versionchanged",
    "deprecated",
]


# Library
# ---------------------------------------------------------------------------
from collections.abc import Callable
from enum import Enum
from functools import partial, wraps
from string import Template
from typing import ClassVar, ParamSpec, TypeVar, overload

# Type
# ---------------------------------------------------------------------------
_P = ParamSpec("_P")  # Parameter type
_R = TypeVar("_R")  # Return type - Can be anything
_T = TypeVar("_T", bound=type)  # Type type - Can be any subtype of `type`

_SPHINX_DOCS_TEMPLATE = Template("$line_break*$mode in version $version$reason*")


# Class
# ---------------------------------------------------------------------------
class SphinxDocstringMode(Enum):
    """
    Enum representing the mode of the version change
    (added, changed, or deprecated)
    """

    ADDED = "Added"
    CHANGED = "Changed"
    DEPRECATED = "Deprecated"


class SphinxDocstring:
    """
    A class-based decorator to add a 'Version added',
    'Version changed', or 'Deprecated' note to a function's docstring,
    formatted for Sphinx documentation.

    Parameters
    ----------
    version : str
        The version in which the function was added, changed, or deprecated.

    reason : str | None, optional
        An optional reason or description for the change
        or deprecation, by default ``None``

    mode : SphinxDocstringMode, optional
        Specifies whether the function was 'added', 'changed', or 'deprecated',
        by default SphinxDocstringMode.ADDED


    Usage
    -----
    Use this as a decorator (``@SphinxDocstring(<parameters>)``)
    """

    _LINEBREAK: ClassVar[str] = "\n\n"  # Use ClassVar for constant

    def __init__(
        self,
        version: str,
        reason: str | None = None,
        mode: SphinxDocstringMode = SphinxDocstringMode.ADDED,
    ) -> None:
        """
        Initializes the SphinxDocstring decorator.

        Parameters
        ----------
        version : str
            The version in which the function was added, changed, or deprecated.

        reason : str | None, optional
            An optional reason or description for the change
            or deprecation, by default ``None``

        mode : SphinxDocstringMode, optional
            Specifies whether the function was 'added', 'changed', or 'deprecated',
            by default SphinxDocstringMode.ADDED

        Usage
        -----
        Use this as a decorator (``@SphinxDocstring(<parameters>)``)
        """
        self.version = version
        self.reason = reason
        self.mode = mode

    @overload  # Class overload
    def __call__(self, obj: _T) -> _T: ...

    @overload  # Function overload
    def __call__(self, obj: Callable[_P, _R]) -> Callable[_P, _R]: ...

    def __call__(self, obj: _T | Callable[_P, _R]) -> _T | Callable[_P, _R]:
        """
        Decorator for class and callable
        """

        # Class wrapper
        if isinstance(obj, type):  # if inspect.isclass(obj):
            return self._update_doc(obj)

        # Function wrapper
        @wraps(obj)
        def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _R:
            return obj(*args, **kwargs)

        return self._update_doc(wrapper)

    def _calculate_white_space(self, docs: str | None) -> int:
        """
        Calculates the number of leading white spaces
        in __doc__ of original function
        """

        res = 0
        if docs is None:
            return res

        try:
            # Replace tabs with space and split line
            lines = docs.expandtabs(4).splitlines()
        except UnicodeError:
            return res
        else:
            # Get indentation of each line and unique it
            indent_set = {len(line) - len(line.lstrip()) for line in lines[1:]}
            # Drop 0
            res_list = sorted([x for x in indent_set if x > 0])

        if res_list:
            return res_list[0]
        return res

    def _generate_version_note(self, num_of_white_spaces: int) -> str:
        """
        Generates the version note string based on the mode
        """
        reason_str = f": {self.reason}" if self.reason else ""
        # return f"{self._LINEBREAK}*{self.mode.value} in version {self.version}{reason_str}*"
        return _SPHINX_DOCS_TEMPLATE.substitute(
            line_break=self._LINEBREAK + " " * num_of_white_spaces,
            mode=self.mode.value,
            version=self.version,
            reason=reason_str,
        )

    @overload
    def _update_doc(self, obj: _T) -> _T: ...
    @overload
    def _update_doc(self, obj: Callable[_P, _R]) -> Callable[_P, _R]: ...
    def _update_doc(self, obj: _T | Callable[_P, _R]) -> _T | Callable[_P, _R]:
        """Update docstring for an object"""
        obj.__doc__ = (obj.__doc__ or "") + self._generate_version_note(
            num_of_white_spaces=self._calculate_white_space(obj.__doc__)
        )
        return obj


# Partial
# ---------------------------------------------------------------------------
versionadded = partial(SphinxDocstring, mode=SphinxDocstringMode.ADDED)
versionchanged = partial(SphinxDocstring, mode=SphinxDocstringMode.CHANGED)
deprecated = partial(SphinxDocstring, mode=SphinxDocstringMode.DEPRECATED)



"""
Test: Package Data

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

import lzma

import pytest

from absfuyu.pkg_data import BasicLZMAOperation, DataList, DataLoader, Pickler


def test_loadData():
    assert True


class TestPickler:
    def test_load(self) -> None:
        Pickler.load(DataList.CHEMISTRY.value)


class TestBasicLZMAOperation:
    def test_load(self) -> None:
        data = "Test"
        compressed = lzma.compress(data.encode(encoding="utf-8"))
        loaded = BasicLZMAOperation.load(compressed).decode()
        assert loaded == data


class TestDataLoader:
    def test_load(self) -> None:
        DataLoader(DataList.CHEMISTRY).load()
        DataLoader(DataList.TAROT).load()
        DataLoader(DataList.PASSWORDLIB).load()
